//****************************************************************************
//**
//**    SND_WIN.CPP
//**    Sound - Windows DirectSound
//**
//**    Copyright (c) 1998 Chris Hargrove
//**    Notice: Usage of any code in this file is subject to the rules
//**            described in the LICENSE.TXT file included in this directory.
//**            Reading, compiling, or otherwise using this code constitutes
//**            automatic acceptance of the rules in said text file.
//**
//****************************************************************************
//============================================================================
//    IMPLEMENTATION HEADERS
//============================================================================
#define SYS_DEFS_INCLUDE_WIN32
#include "sys_defs.h"
#include "snd_win.h"

#include <dsound.h>

//============================================================================
//    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
//============================================================================
#define SNDW_MAXSOUNDHANDLES		256
#define SNDW_MAXPLAYHANDLES			2048

#define SNDW_SAMPLERATE_DEFAULT			11025
#define SNDW_SAMPLEBITDEPTH_DEFAULT		8
#define SNDW_SAMPLECHANNELS_DEFAULT		1

enum
{
	PIF_LOOPING = 0x00000001,
	PIF_PAUSED  = 0x00000002
};

//============================================================================
//    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
//============================================================================
typedef struct sndw_playInfo_s sndw_playInfo_t;
typedef struct sndw_soundInfo_s sndw_soundInfo_t;

struct sndw_playInfo_s
{
	LPDIRECTSOUNDBUFFER dsDuplicateBuffer; // sound buffer duplicate, NULL if not in use
	sndw_playInfo_t *prev, *next; // sibling links, self-linked if not in use
	sndw_soundInfo_t* sound; // link to parent sound, NULL if not in use
	dword flags; // PIF_ flags
};

struct sndw_soundInfo_s
{
	LPDIRECTSOUNDBUFFER dsSoundBuffer; // actual sound buffer, NULL if sound is not in use
	sndw_playInfo_t playList; // dummy list headers for play handles
	float baseFrequency; // base frequency (sample rate) of sound
};

//============================================================================
//    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE DATA
//============================================================================

// directsound stuff
static LPDIRECTSOUND sndw_lpDS=NULL;
static LPDIRECTSOUNDBUFFER sndw_lpdsbPrimary=NULL;
static DSCAPS snd_dsCaps;

// handle info pools
static sndw_soundInfo_t sndw_soundPool[SNDW_MAXSOUNDHANDLES];
static sndw_playInfo_t sndw_playPool[SNDW_MAXPLAYHANDLES];
static sndw_playInfo_t sndw_playPoolFreeList;

//============================================================================
//    INTERFACE DATA
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE FUNCTIONS
//============================================================================
static sndw_soundInfo_t* SoundInfoForId(dword id)
{
	if ((!id) || (id > SNDW_MAXSOUNDHANDLES))
		return(NULL);
	if (!sndw_soundPool[id-1].dsSoundBuffer)
		return(NULL);
	return(&sndw_soundPool[id-1]);
}

static sndw_playInfo_t* PlayInfoForId(dword id)
{
	if ((!id) || (id > SNDW_MAXPLAYHANDLES))
		return(NULL);
	if (!sndw_playPool[id-1].dsDuplicateBuffer)
		return(NULL);
	return(&sndw_playPool[id-1]);
}

static dword IdForSoundInfo(sndw_soundInfo_t* si)
{
	if (!si)
		return(0);
	return((dword)(si - sndw_soundPool)+1);
}

static dword IdForPlayInfo(sndw_playInfo_t* pi)
{
	if (!pi)
		return(0);
	return((dword)(pi - sndw_playPool)+1);
}

static bbool CreateBuffer(LPDIRECTSOUNDBUFFER* outLpdsb, dword bufferSize,
						  dword sampleRate, dword bitDepth, dword channels)
{
	WAVEFORMATEX waveFmt;
	DSBUFFERDESC dsBufDesc;

	memset(&waveFmt, 0, sizeof(WAVEFORMATEX));
	waveFmt.wFormatTag = WAVE_FORMAT_PCM;
	waveFmt.nSamplesPerSec = sampleRate;
	waveFmt.wBitsPerSample = bitDepth;
	waveFmt.nChannels = channels;
	waveFmt.nBlockAlign = (channels*bitDepth)>>3;
	waveFmt.nAvgBytesPerSec = sampleRate*waveFmt.nBlockAlign;

	memset(&dsBufDesc, 0, sizeof(DSBUFFERDESC));
	dsBufDesc.dwSize = sizeof(DSBUFFERDESC);
	dsBufDesc.dwFlags = DSBCAPS_CTRLDEFAULT | DSBCAPS_STATIC;
	dsBufDesc.dwBufferBytes = bufferSize;
	dsBufDesc.lpwfxFormat = &waveFmt;

	if (sndw_lpDS->CreateSoundBuffer(&dsBufDesc, outLpdsb, NULL) != DS_OK)
		return(0);
	
	return(1);
}

//============================================================================
//    INTERFACE FUNCTIONS
//============================================================================

// init/shutdown

void SNDW_Init()
{
	dword i;
	DSBUFFERDESC dsBufDesc;
	
	if (DirectSoundCreate(NULL, &sndw_lpDS, NULL) != DS_OK)
		SYS_Errorf("SNDW_Init failure: DirectSoundCreate");
	if (sndw_lpDS->SetCooperativeLevel(sysw_hwnd, DSSCL_PRIORITY) != DS_OK)
		SYS_Errorf("SNDW_Init failure: SetCooperativeLevel");
	snd_dsCaps.dwSize = sizeof(DSCAPS);
	if (sndw_lpDS->GetCaps(&snd_dsCaps) != DS_OK)
		SYS_Errorf("SNDW_Init failure: GetCaps");

	memset(&dsBufDesc, 0, sizeof(DSBUFFERDESC));
	dsBufDesc.dwSize = sizeof(DSBUFFERDESC);
	dsBufDesc.dwFlags |= DSBCAPS_PRIMARYBUFFER | DSBCAPS_CTRLVOLUME;
	dsBufDesc.dwBufferBytes = 0;
	dsBufDesc.lpwfxFormat = NULL;
	if (sndw_lpDS->CreateSoundBuffer(&dsBufDesc, &sndw_lpdsbPrimary, NULL) != DS_OK)
		SYS_Errorf("SNDW_Init failure: CreateSoundBuffer (primary)");

	// add play handles to global free list
	sndw_playPoolFreeList.next = sndw_playPoolFreeList.prev = &sndw_playPoolFreeList;
	for (i=0;i<SNDW_MAXPLAYHANDLES;i++)
	{
		sndw_playInfo_t* pi = &sndw_playPool[i];
		pi->next = sndw_playPoolFreeList.next;
		pi->prev = &sndw_playPoolFreeList;
		pi->next->prev = pi->prev->next = pi;
	}
}

void SNDW_Shutdown()
{
	dword i;

	// free all sounds
	for (i=1;i<=SNDW_MAXSOUNDHANDLES;i++)
		SNDW_FreeSound(i);

	// kill the dsound object
	if (sndw_lpDS)
	{
		sndw_lpDS->Release();
		sndw_lpDS = NULL;
	}
}

// sound control

dword SNDW_LoadSound(dword sampleRate, dword bitDepth, dword channels,
					 dword sampleSizeBytes, byte* sampleData, dword maxPlayHandles)
{
	dword i;
	HRESULT hr;
	sndw_soundInfo_t* si=NULL;
	void *lockPtr1, *lockPtr2;
	dword lockSize1, lockSize2;

	// find a free sound handle
	for (i=0;i<SNDW_MAXSOUNDHANDLES;i++)
	{
		si = &sndw_soundPool[i];
		if (!si->dsSoundBuffer)
			break;
	}
	if (i==SNDW_MAXSOUNDHANDLES)
		return(0); // no more sound handles
	if (!CreateBuffer(&si->dsSoundBuffer, sampleSizeBytes, sampleRate, bitDepth, channels))
		return(0); // couldn't create sound buffer
	
	// lock the buffer
	if ((hr = si->dsSoundBuffer->Lock(0, sampleSizeBytes, &lockPtr1, &lockSize1, &lockPtr2, &lockSize2, 0)) != DS_OK)
	{
		if (hr != DSERR_BUFFERLOST)
			return(0);
		si->dsSoundBuffer->Restore();
		if (si->dsSoundBuffer->Lock(0, sampleSizeBytes, &lockPtr1, &lockSize1, &lockPtr2, &lockSize2, 0) != DS_OK)
			return(0);
	}

	// write the data
	if (lockSize1)
		memcpy(lockPtr1, sampleData, lockSize1);
	if (lockSize2)
		memcpy(lockPtr2, sampleData+lockSize1, lockSize2);

	// unlock it
	si->dsSoundBuffer->Unlock(lockPtr1, lockSize1, lockPtr2, lockSize2);

	// allocate play handles
	si->playList.next = si->playList.prev = &si->playList;
	for (i=0;i<maxPlayHandles;i++)
	{
		sndw_playInfo_t* pi = sndw_playPoolFreeList.next;
		if (pi == &sndw_playPoolFreeList)
			continue;
		pi->sound = si;
		pi->flags = 0;
		pi->next->prev = pi->prev;
		pi->prev->next = pi->next;
		pi->next = si->playList.next;
		pi->prev = &si->playList;
		pi->next->prev = pi->prev->next = pi;
		if (sndw_lpDS->DuplicateSoundBuffer(si->dsSoundBuffer, &pi->dsDuplicateBuffer) != DS_OK)
			SYS_Errorf("SNDW_LoadSound: DuplicateSoundBuffer failure");
	}

	si->baseFrequency = sampleRate;

	return(IdForSoundInfo(si));
}

void SNDW_FreeSound(dword soundHandle)
{
	sndw_soundInfo_t* si;

	if (!(si = SoundInfoForId(soundHandle)))
		return;
	
	// free all playhandles back to the global free list
	while (si->playList.next != &si->playList)
	{
		sndw_playInfo_t* pi = si->playList.next;
		if (SNDW_IsSoundPlaying(IdForPlayInfo(pi)))
			SNDW_StopSound(IdForPlayInfo(pi));
		pi->next->prev = pi->prev;
		pi->prev->next = pi->next;
		pi->next = sndw_playPoolFreeList.next;
		pi->prev = &sndw_playPoolFreeList;
		pi->next->prev = pi->prev->next = pi;
		pi->sound = NULL;
		pi->flags = 0;
		if (pi->dsDuplicateBuffer)
		{
			pi->dsDuplicateBuffer->Release();
			pi->dsDuplicateBuffer = NULL;
		}
	}
	
	// release the sound buffer
	si->dsSoundBuffer->Release();
	si->dsSoundBuffer = NULL;
}

// play control

dword SNDW_PlaySound(dword soundHandle, bbool looping)
{
	HRESULT hr;
	sndw_soundInfo_t* si;
	sndw_playInfo_t* pi;
	dword flags=0;

	if (!(si = SoundInfoForId(soundHandle)))
		return(0);
	for (pi=si->playList.next; pi!=&si->playList; pi=pi->next)
	{
		if (!SNDW_IsSoundPlaying(IdForPlayInfo(pi)))
			break;
	}
	if (pi==&si->playList)
		return(0);
	pi->flags = 0;
	if (looping)
		flags |= DSBPLAY_LOOPING;
	if (pi->dsDuplicateBuffer->SetCurrentPosition(0) != DS_OK)
		return(0);
	if ((hr = pi->dsDuplicateBuffer->Play(0, 0, flags)) != DS_OK)
	{
		if (hr != DSERR_BUFFERLOST)
			return(0);
		pi->dsDuplicateBuffer->Restore();
		if (pi->dsDuplicateBuffer->Play(0, 0, flags) != DS_OK)
			return(0);
	}
	if (looping)
		pi->flags |= PIF_LOOPING;
	return(IdForPlayInfo(pi));
}

void SNDW_StopSound(dword playHandle)
{
	sndw_playInfo_t* pi;

	if (!(pi = PlayInfoForId(playHandle)))
		return;
	pi->dsDuplicateBuffer->Stop();
}

void SNDW_PauseSound(dword playHandle)
{
	sndw_playInfo_t* pi;

	if (!(pi = PlayInfoForId(playHandle)))
		return;
	pi->dsDuplicateBuffer->Stop();
	pi->flags |= PIF_PAUSED;
}

void SNDW_ResumeSound(dword playHandle)
{
	HRESULT hr;
	sndw_playInfo_t* pi;
	dword flags;

	if (!(pi = PlayInfoForId(playHandle)))
		return;
	if (!(pi->flags & PIF_PAUSED))
		return;
	if (pi->flags & PIF_LOOPING)
		flags |= DSBPLAY_LOOPING;
	if ((hr = pi->dsDuplicateBuffer->Play(0, 0, flags)) != DS_OK)
	{
		if (hr != DSERR_BUFFERLOST)
			return;
		pi->dsDuplicateBuffer->Restore();
		if (pi->dsDuplicateBuffer->Play(0, 0, flags) != DS_OK)
			return;
	}
	pi->flags &= ~PIF_PAUSED;
}

void SNDW_SetSoundVolume(dword playHandle, float volume)
{
	sndw_playInfo_t* pi;

	if (!(pi = PlayInfoForId(playHandle)))
		return;
	LONG dsVolume = -1000.0*(log(1.0/volume) / log(2.0));
	if (dsVolume > 0)
		dsVolume = 0;
	if (dsVolume < -10000)
		dsVolume = -10000; // directsound limit, corresponds to volume of 1/1024, effectively silent
	pi->dsDuplicateBuffer->SetVolume(dsVolume);
}

void SNDW_SetSoundPan(dword playHandle, float pan)
{
	sndw_playInfo_t* pi;
	LONG dsPan;

	if (!(pi = PlayInfoForId(playHandle)))
		return;
	if (pan < 0.0)
	{
		dsPan = -1000.0*(log(1.0/(1.0+pan)) / log(2.0));
		if (dsPan > 0)
			dsPan = 0;
		if (dsPan < -10000)
			dsPan = -10000;
	}
	else
	{
		dsPan = 1000.0*(log(1.0/(1.0-pan)) / log(2.0));
		if (dsPan < 0)
			dsPan = 0;
		if (dsPan > 10000)
			dsPan = 10000;
	}
	pi->dsDuplicateBuffer->SetPan(dsPan);
}

void SNDW_SetSoundFrequency(dword playHandle, float freq)
{
	sndw_playInfo_t* pi;
	DWORD dsFreq;

	if (!(pi = PlayInfoForId(playHandle)))
		return;
	dsFreq = pi->sound->baseFrequency*freq;
	if (dsFreq < 80)
		dsFreq = 80;
	if (dsFreq > 50000)
		dsFreq = 50000;
	pi->dsDuplicateBuffer->SetFrequency(dsFreq);
}

float SNDW_GetSoundFrequency(dword playHandle)
{
	sndw_playInfo_t* pi;
	DWORD dsFreq;

	if (!(pi = PlayInfoForId(playHandle)))
		return(0.0);
	if (pi->dsDuplicateBuffer->GetFrequency(&dsFreq) != DS_OK)
		return(1.0);
	if (dsFreq < 80)
		dsFreq = 80;
	if (dsFreq > 50000)
		dsFreq = 50000;
	return((float)dsFreq / pi->sound->baseFrequency);
}

bbool SNDW_IsSoundPlaying(dword playHandle)
{
	sndw_playInfo_t* pi;
	DWORD statflags;

	if (!(pi = PlayInfoForId(playHandle)))
		return(0);
	if (pi->dsDuplicateBuffer->GetStatus(&statflags) != DS_OK)
		return(0);
	return(statflags & DSBSTATUS_PLAYING);
}

// misc operations

void SNDW_SetGlobalVolume(float volume)
{
	if (!sndw_lpdsbPrimary)
		return;
	LONG dsVolume = -1000.0*(log(1.0/volume) / log(2.0));
	if (dsVolume > 0)
		dsVolume = 0;
	if (dsVolume < -10000)
		dsVolume = -10000;
	sndw_lpdsbPrimary->SetVolume(dsVolume);
}

float SNDW_GetGlobalVolume()
{
	LONG dsVolume;
	float attenuation;

	if (!sndw_lpdsbPrimary)
		return(0.0);
	sndw_lpdsbPrimary->GetVolume(&dsVolume);
	attenuation = 1.0 / pow(2.718281829,(-0.001*(float)dsVolume)*log(2.0));
	if (attenuation < 1.0/1024.0)
		attenuation = 1.0/1024.0;
	if (attenuation > 1.0)
		attenuation = 1.0;
	return(attenuation);
}

//============================================================================
//    INTERFACE CLASS BODIES
//============================================================================

//****************************************************************************
//**
//**    END MODULE SND_WIN.CPP
//**
//****************************************************************************

