//****************************************************************************
//**
//**    SYS_MAIN.CPP
//**    System Main
//**
//**    Copyright (c) 1998 Chris Hargrove
//**    Notice: Usage of any code in this file is subject to the rules
//**            described in the LICENSE.TXT file included in this directory.
//**            Reading, compiling, or otherwise using this code constitutes
//**            automatic acceptance of the rules in said text file.
//**
//****************************************************************************
//============================================================================
//    IMPLEMENTATION HEADERS
//============================================================================
#define SYS_DEFS_INCLUDE_WIN32 // for SYSW_Argc, SYSW_Argv,
                               // and SYSW_MessageBox
#include "sys_defs.h"
#include "sys_main.h"
//============================================================================
//    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
//============================================================================
#define SAFEMALLOC_MARKER 0x2b3d1234
#define SAFEMALLOC_USEMARKER // enable malloc/free marker verify

//============================================================================
//    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
//============================================================================
//============================================================================
//    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE DATA
//============================================================================
static float sys_baseTime;
static float sys_lastTime;

static dword testImages[2];
static res_id_t testImageResIDs[2];

//============================================================================
//    INTERFACE DATA
//============================================================================
char sys_programPath[SYS_MAX_PATH];
float sys_curTime, sys_frameTime;

//============================================================================
//    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
//============================================================================
static void UpdateFrameTimes();
static void ResetFrameTimes();

//============================================================================
//    IMPLEMENTATION PRIVATE FUNCTIONS
//============================================================================
static void UpdateFrameTimes()
{
	sys_curTime = M_TimeFloat() - sys_baseTime;
	sys_frameTime = sys_curTime - sys_lastTime;
	sys_lastTime = sys_curTime;
}

static void ResetFrameTimes()
{
	sys_baseTime = M_TimeFloat();
	sys_lastTime = sys_curTime = sys_frameTime = 0.0;
}

static bbool ImageReceiver(in_event_t* event)
{
    if ((event->eventType == INEV_PRESS) && (event->key == INKEY_ESCAPE))
        SYS_Quit();
    return(0);
}

//============================================================================
//    INTERFACE FUNCTIONS
//============================================================================

// The "Big Three" control functions

// initialization
void SYS_Init()
{
    {
        // vg_vec.h test, doesn't need anything more than this message box
        vgvec3 v(100,0,0); // point 100 units along X axis, in parent space
        vgocs3 ocs(
            // ocs origin 500 units along both X and Y
            vgvec3(500,500,0),
            // ocs axial frame is the cartesian axes rotated about the Y axis PI/2 radians (90 degrees)
            vgframe3(vgvec3(0,1,0), VG_PI*0.5),
            // ocs has 2x scale axes (so anything relative to it will be half the size)
            vgvec3(2,2,2)
            );

        vgvec3 vTrans = v >> ocs; // v from the parent space, relative to the child OCS

        SYS_Alertf("Hi!  I'm an example of the new vector geometry classes.\n"
                   "Draw this example out on paper if you need help visualizing it.\n\n"
                   "Point in parent space: (%.4f, %.4f, %.4f)\n\n"
                   "Child OCS:\n"
                   "\tOrigin: (%.4f, %.4f, %.4f)\n"
                   "\tAxial Frame X: (%.4f, %.4f, %.4f)\n"
                   "\tAxial Frame Y: (%.4f, %.4f, %.4f)\n"
                   "\tAxial Frame Z: (%.4f, %.4f, %.4f)\n"
                   "\tAxis Scales: (%.4f, %.4f, %.4f)\n\n"
                   "Point relative to OCS: (%.4f, %.4f, %.4f)\n",
                   v.x, v.y, v.z,
                   ocs.translate.x, ocs.translate.y, ocs.translate.z,
                   ocs.frame.vX.x, ocs.frame.vX.y, ocs.frame.vX.z,
                   ocs.frame.vY.x, ocs.frame.vY.y, ocs.frame.vY.z,
                   ocs.frame.vZ.x, ocs.frame.vZ.y, ocs.frame.vZ.z,
                   ocs.scale.x, ocs.scale.y, ocs.scale.z,
                   vTrans.x, vTrans.y, vTrans.z);

        exit(0); // don't use SYS_Quit() since we're not actually starting up with this example
	}


    MEM_Init(); // initialize memory system
	XF_Init(); // initialize file system
    RES_Init(); // initialize resource manager
	IN_Init(); // initialize user input
    CON_Init(); // initialize console
    SND_Init(); // initialize sound system    
    VID_Init(); // initialize video system

	ResetFrameTimes(); // initialize frame times from this point
}

// main loop
void SYS_Frame()
{
	UpdateFrameTimes();
	
	RES_Frame(); // update resources
    IN_Frame(); // process user input
    VID_Frame(); // video buffer update
}

// de-initialization
void SYS_Shutdown()
{	
    static bbool shuttingDown = 0;
    if (shuttingDown)
        return; // recursive shutdown block
    shuttingDown = 1;

    VID_Shutdown();
    SND_Shutdown();
    CON_Shutdown();
    IN_Shutdown();
    RES_Shutdown();
    XF_Shutdown();
    MEM_Shutdown();
}

// alert notification
void SYS_Alertf(char* text, ...)
{
	static char textBuffer[2048];
	va_list args;
	va_start(args, text);
	vsprintf(textBuffer, text, args);
	va_end(args);
	SYSW_MessageBox(SYS_APPTITLE " Notification", MB_OK, textBuffer);
}

// initialize debugging window
void SYS_InitDebug()
{
	SYSW_InitDebug();
}

// shut down debugging window
void SYS_ShutdownDebug()
{
	SYSW_ShutdownDebug();
}

// debugging window output
void SYS_Debugf(char* text, ...)
{
	static char textBuffer[1024];
	va_list args;
	va_start(args, text);
	vsprintf(textBuffer, text, args);
	va_end(args);
	SYSW_DebugString(textBuffer);
}

// exit without error
void SYS_Quit()
{
	SYS_Shutdown();
	exit(0);
}

// exit with fatal error
void SYS_Errorf(char *text, ...)
{
	static char textBuffer[1024];
	va_list args;
	va_start(args, text);
	vsprintf(textBuffer, text, args);
	va_end(args);
	SYSW_MessageBox("Fatal Error", MB_OK, textBuffer);
	SYS_Shutdown();
	exit(1);
}

// command line argument functions, call windows equivalents
byte SYS_Argc()
{
	return(SYSW_Argc());
}

char* SYS_Argv(byte index)
{
	return(SYSW_Argv(index));
}

// safe malloc and free wrappers, with optional marker verify (4 byte cost)
void* SYS_SafeMalloc(int size)
{
#ifdef SAFEMALLOC_USEMARKER	
	size += 4;
#endif
	void *res = malloc(size);
	if (!res)
		SYS_Errorf("SYS_SafeMalloc: Out of memory on %d bytes", size);
#ifdef SAFEMALLOC_USEMARKER	
	*((long *)res) = SAFEMALLOC_MARKER;
	return((long *)res+1);
#else
	return(res);
#endif
}

void SYS_SafeFree(void **ptr)
{
	long *p;
	if (!(*ptr))
		SYS_Errorf("SYS_SafeFree: Attempt to free null pointer");
	p = (long *)*ptr;
#ifdef SAFEMALLOC_USEMARKER	
	p -= 1;
	if (*p != SAFEMALLOC_MARKER)
    {
		if (*p == (SAFEMALLOC_MARKER ^ 0xFFFFFFFF))
            SYS_Errorf("SYS_SafeFree: Attempt to free pointer twice");
        else
            SYS_Errorf("SYS_SafeFree: Attempt to free non-SafeMalloc'ed pointer");
    }
    *p = SAFEMALLOC_MARKER ^ 0xFFFFFFFF;
#endif
	free(p);
	*ptr = NULL;
}

//============================================================================
//    INTERFACE CLASS BODIES
//============================================================================

//****************************************************************************
//**
//**    END MODULE SYS_MAIN.CPP
//**
//****************************************************************************

