#ifndef __VG_VEC_H__
#define __VG_VEC_H__
//****************************************************************************
//**
//**    VG_VEC.H
//**    Header - Vector Geometry
//**
//**    Copyright (c) 1998/1999 Chris Hargrove
//**    Notice: Usage of any code in this file is subject to the rules
//**            described in the LICENSE.TXT file included in this directory.
//**            Reading, compiling, or otherwise using this code constitutes
//**            automatic acceptance of the rules in said text file.
//**
//****************************************************************************
//============================================================================
//    INTERFACE REQUIRED HEADERS
//============================================================================
// don't bother including any other system files; these are all that's needed
#include <math.h>
#include <float.h>

//============================================================================
//    INTERFACE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
//============================================================================
#define VG_PI 3.1415926535897932385 // pi
#define VG_EPSILON (1E-3) // epsilon "small value"
#define VG_EPSILON2 (1E-6) // squared epsilon

#pragma warning(disable:4244 4305) // conversion warnings

//============================================================================
//    INTERFACE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
//============================================================================
class vgvec2; // 2D vector/point
class vgvec3; // 3D vector/point
class vgframe3; // 3D axial frame
class vgocs3; // 3D orthogonal coordinate system

//============================================================================
//    INTERFACE STRUCTURES / UTILITY CLASSES
//============================================================================
//============================================================================
//    INTERFACE DATA DECLARATIONS
//============================================================================
//============================================================================
//    INTERFACE FUNCTION PROTOTYPES
//============================================================================

// Quicker absolute value than fabs()
// Takes advantage of the assumption that the float is IEEE 32-bit (s1.e8.m23)
// If you're porting this code, you may want to alter this to use fabs() instead.
inline float VG_Fabs(float f)
{
    return(*((unsigned long *)(&f)) & 0x7fffffff);
}

//============================================================================
//    INTERFACE OBJECT CLASS DEFINITIONS
//============================================================================

//----------------------------------------
//
// vgvec2
// 2D vector/point
//
//----------------------------------------

class vgvec2
{
public:
	float x, y;

	vgvec2(); // default constructor, no initialization
	vgvec2(const vgvec2& inV); // copy constructor
	vgvec2(float inX, float inY); // component constructor

	vgvec2& operator = (const vgvec2& inV); // assignment
	vgvec2& operator += (const vgvec2& inV); // component add assignment
	vgvec2& operator -= (const vgvec2& inV); // component sub assignment
	vgvec2& operator *= (const vgvec2& inV); // component mul assignment
	vgvec2& operator /= (const vgvec2& inV); // component div assignment
	vgvec2& operator *= (float inScale); // scalar mul assignment
	vgvec2& operator /= (float inScale); // scalar div assignment

	operator float* (); // cast to float*
	
	float Length2() const; // squared length
	float Length() const; // length of vector (distance of point from origin)
	float Normalize(); // normalize vector, return length
	int Dominant() const; // dominant axis index

	friend vgvec2 operator - (const vgvec2& inV); // unary component negate
	friend vgvec2 operator + (const vgvec2& inV1, const vgvec2& inV2); // component add
	friend vgvec2 operator - (const vgvec2& inV1, const vgvec2& inV2); // component sub
	friend vgvec2 operator * (const vgvec2& inV1, const vgvec2& inV2); // component mul
	friend vgvec2 operator / (const vgvec2& inV1, const vgvec2& inV2); // component div
	friend vgvec2 operator * (const vgvec2& inV, float inScale); // scalar mul
	friend vgvec2 operator / (const vgvec2& inV, float inScale); // scalar div
	friend float operator & (const vgvec2& inV1, const vgvec2& inV2); // distance between points
	friend float operator | (const vgvec2& inV1, const vgvec2& inV2); // dot product
	friend vgvec2 operator ~ (const vgvec2& inV); // unary normal
	friend bool operator ! (const vgvec2& inV); // epsilon degeneracy, near-origin point
	friend bool operator == (const vgvec2& inV1, const vgvec2& inV2); // epsilon length equality
	friend bool operator != (const vgvec2& inV1, const vgvec2& inV2); // epsilon length inequality
	friend bool operator <= (const vgvec2& inV1, const vgvec2& inV2); // epsilon length less or equal
	friend bool operator >= (const vgvec2& inV1, const vgvec2& inV2); // epsilon length greater or equal
	friend bool operator < (const vgvec2& inV1, const vgvec2& inV2); // epsilon length less than
	friend bool operator > (const vgvec2& inV1, const vgvec2& inV2); // epsilon length greater than
};


//----------------------------------------
//
// vgvec3
// 3D vector/point
//
//----------------------------------------

class vgvec3
{
public:
	float x, y, z;

	vgvec3(); // default constructor, no initialization
	vgvec3(const vgvec3& inV); // copy constructor
	vgvec3(float inX, float inY, float inZ); // component constructor
	vgvec3(const vgvec2& inV); // 2D vector/point upgrade constructor

	vgvec3& operator = (const vgvec3& inV); // assignment
	vgvec3& operator += (const vgvec3& inV); // component add assignment
	vgvec3& operator -= (const vgvec3& inV); // component sub assignment
	vgvec3& operator *= (const vgvec3& inV); // component mul assignment
	vgvec3& operator /= (const vgvec3& inV); // component div assignment
	vgvec3& operator *= (float inScale); // scalar mul assignment
	vgvec3& operator /= (float inScale); // scalar div assignment
	vgvec3& operator >>= (const vgframe3& inF); // axial frame inbound transform
	vgvec3& operator <<= (const vgframe3& inF); // axial frame outbound transform
	vgvec3& operator >>= (const vgocs3& inOCS); // OCS inbound transform
	vgvec3& operator <<= (const vgocs3& inOCS); // OCS outbound transform

	operator float* (); // cast to float*

	float Length2() const; // squared length
	float Length() const; // length of vector (distance of point from origin)
	float Normalize(); // normalize vector, return length
	int Dominant(); // dominant axis index

	friend vgvec3 operator - (const vgvec3& inV); // unary component negate
	friend vgvec3 operator + (const vgvec3& inV1, const vgvec3& inV2); // component add
	friend vgvec3 operator - (const vgvec3& inV1, const vgvec3& inV2); // component sub
	friend vgvec3 operator * (const vgvec3& inV1, const vgvec3& inV2); // component mul
	friend vgvec3 operator / (const vgvec3& inV1, const vgvec3& inV2); // component div
	friend vgvec3 operator * (const vgvec3& inV, float inScale); // scalar mul
	friend vgvec3 operator / (const vgvec3& inV, float inScale); // scalar div
	friend float operator & (const vgvec3& inV1, const vgvec3& inV2); // distance between points
	friend float operator | (const vgvec3& inV1, const vgvec3& inV2); // dot product
	friend vgvec3 operator ^ (const vgvec3& inV1, const vgvec3& inV2); // cross product
	friend bool operator ! (const vgvec3& inV); // epsilon degeneracy, near-origin point
	friend bool operator == (const vgvec3& inV1, const vgvec3& inV2); // epsilon length equality
	friend bool operator != (const vgvec3& inV1, const vgvec3& inV2); // epsilon length inequality
	friend bool operator <= (const vgvec3& inV1, const vgvec3& inV2); // epsilon length less or equal
	friend bool operator >= (const vgvec3& inV1, const vgvec3& inV2); // epsilon length greater or equal
	friend bool operator < (const vgvec3& inV1, const vgvec3& inV2); // epsilon length less than
	friend bool operator > (const vgvec3& inV1, const vgvec3& inV2); // epsilon length greater than
};


//----------------------------------------
//
// vgframe3
// 3D axial frame
//
//----------------------------------------

class vgframe3
{
public:
	vgvec3 vX, vY, vZ; // local normalized X, Y, and Z axes, right-handed

	vgframe3(); // default constructor, initialized to cartesian axes
	vgframe3(const vgframe3& inF); // copy constructor
	vgframe3(const vgvec3& inX, const vgvec3& inY, const vgvec3& inZ); // component constructor
    vgframe3(const vgvec3& inAxis, float inRadians, bool normalizeAxis=1); // axis and angle rotation constructor

	vgframe3& operator = (const vgframe3& inF); // assignment
	vgframe3& operator >>= (const vgframe3& inF); // axial frame inbound transform assignment
	vgframe3& operator <<= (const vgframe3& inF); // axial frame outbound transform assignment

	friend vgvec3 operator >> (const vgvec3& inV, const vgframe3& inF); // vgvec3 inbound transform to axial frame
	friend vgvec3 operator << (const vgvec3& inV, const vgframe3& inF); // vgvec3 outbound transform from axial frame
	friend vgframe3 operator >> (const vgframe3& inF1, const vgframe3& inF2); // axial frame inbound transform
	friend vgframe3 operator << (const vgframe3& inF1, const vgframe3& inF2); // axial frame outbound transform
};


//----------------------------------------
//
// vgocs3
// 3D orthogonal coordinate system
//
//----------------------------------------

class vgocs3
{
public:	
	vgvec3 translate; // OCS origin
	vgframe3 frame; // OCS axial frame
	vgvec3 scale; // OCS axis scales

	vgocs3(); // default constructor, zero translation with cartesian axes and unit scale
	vgocs3(const vgocs3& inOCS); // copy constructor
	vgocs3(const vgvec3& inTranslate, const vgframe3& inFrame, const vgvec3& inScale); // component constructor

	vgocs3& operator = (const vgocs3& inOCS); // assignment
	vgocs3& operator += (const vgvec3& inTranslate); // translation component add assignment
	vgocs3& operator -= (const vgvec3& inTranslate); // translation component sub assignment
	vgocs3& operator *= (const vgvec3& inScale); // scale component mul assignment
	vgocs3& operator /= (const vgvec3& inScale); // scale component div assignment
	vgocs3& operator *= (float inScale); // scale scalar mul assignment
	vgocs3& operator /= (float inScale); // scale scalar div assignment
	vgocs3& operator >>= (const vgocs3& inOCS); // OCS inbound transform assignment
	vgocs3& operator <<= (const vgocs3& inOCS); // OCS outbound transform assignment

	friend vgvec3 operator >> (const vgvec3& inV, const vgocs3& inOCS); // vgvec3 inbound transform to OCS
	friend vgvec3 operator << (const vgvec3& inV, const vgocs3& inOCS); // vgvec3 outbound transform from OCS
	friend vgocs3 operator >> (const vgocs3& inOCS1, const vgocs3& inOCS2); // OCS inbound transform
	friend vgocs3 operator << (const vgocs3& inOCS1, const vgocs3& inOCS2); // OCS outbound transform
};

//============================================================================
//    INTERFACE CLASS BODIES
//============================================================================
#define VG_CALL inline
// If you're using MSVC6, you can comment out the above line and use the one
// below instead.
//#define VG_CALL __forceinline

// All of the class bodies listed are declared here inline for better
// performance.  If you don't want to inline them (for some reason), you can
// define VG_CALL to a blank and move everything below here to an actual
// .cpp file instead.

//----------------------------------------
//
// vgvec2
// 2D vector/point
//
//----------------------------------------

VG_CALL vgvec2::vgvec2() {}
VG_CALL vgvec2::vgvec2(const vgvec2& inV) { x = inV.x; y = inV.y; }
VG_CALL vgvec2::vgvec2(float inX, float inY) { x = inX; y = inY; }

VG_CALL vgvec2& vgvec2::operator = (const vgvec2& inV) { x = inV.x; y = inV.y; return(*this); }
VG_CALL vgvec2& vgvec2::operator += (const vgvec2& inV) { x += inV.x; y += inV.y; return(*this); }
VG_CALL vgvec2& vgvec2::operator -= (const vgvec2& inV) { x -= inV.x; y -= inV.y; return(*this); }
VG_CALL vgvec2& vgvec2::operator *= (const vgvec2& inV) { x *= inV.x; y *= inV.y; return(*this); }
VG_CALL vgvec2& vgvec2::operator /= (const vgvec2& inV) { x /= inV.x; y /= inV.y; return(*this); }
VG_CALL vgvec2& vgvec2::operator *= (float inScale) { x *= inScale; y *= inScale; return(*this); }
VG_CALL vgvec2& vgvec2::operator /= (float inScale) { x /= inScale; y /= inScale; return(*this); }

VG_CALL vgvec2::operator float* () { return((float*)this); }

VG_CALL float vgvec2::Length2() const { return(x*x + y*y); }
VG_CALL float vgvec2::Length() const { return(sqrt(x*x + y*y)); }
VG_CALL float vgvec2::Normalize() { float a(Length()); float b(1.0/a); x *= b; y *= b; return(a); }
VG_CALL int vgvec2::Dominant() const { return(VG_Fabs(y) > VG_Fabs(x)); }

VG_CALL vgvec2 operator - (const vgvec2& inV) { return(vgvec2(-inV.x, -inV.y)); }
VG_CALL vgvec2 operator + (const vgvec2& inV1, const vgvec2& inV2) { return(vgvec2(inV1.x+inV2.x, inV1.y+inV2.y)); }
VG_CALL vgvec2 operator - (const vgvec2& inV1, const vgvec2& inV2) { return(vgvec2(inV1.x-inV2.x, inV1.y-inV2.y)); }
VG_CALL vgvec2 operator * (const vgvec2& inV1, const vgvec2& inV2) { return(vgvec2(inV1.x*inV2.x, inV1.y*inV2.y)); }
VG_CALL vgvec2 operator / (const vgvec2& inV1, const vgvec2& inV2) { return(vgvec2(inV1.x/inV2.x, inV1.y/inV2.y)); }
VG_CALL vgvec2 operator * (const vgvec2& inV, float inScale) { return(vgvec2(inV.x*inScale, inV.y*inScale)); }
VG_CALL vgvec2 operator / (const vgvec2& inV, float inScale) { return(vgvec2(inV.x/inScale, inV.y/inScale)); }
VG_CALL float operator & (const vgvec2& inV1, const vgvec2& inV2) { return((inV2 - inV1).Length()); }
VG_CALL float operator | (const vgvec2& inV1, const vgvec2& inV2) { return(inV1.x*inV2.x + inV1.y*inV2.y); }
VG_CALL vgvec2 operator ~ (const vgvec2& inV) { return(vgvec2(inV.y, -inV.x)); }
VG_CALL bool operator ! (const vgvec2& inV) { return(inV.Length2() <= VG_EPSILON2); }
VG_CALL bool operator == (const vgvec2& inV1, const vgvec2& inV2) { return((inV2-inV1).Length2() <= VG_EPSILON2); }
VG_CALL bool operator != (const vgvec2& inV1, const vgvec2& inV2) { return((inV2-inV1).Length2() > VG_EPSILON2); }
VG_CALL bool operator <= (const vgvec2& inV1, const vgvec2& inV2) { return(inV1.Length2() <= (inV2.Length2()+VG_EPSILON2)); }
VG_CALL bool operator >= (const vgvec2& inV1, const vgvec2& inV2) { return(inV1.Length2() >= (inV2.Length2()-VG_EPSILON2)); }
VG_CALL bool operator < (const vgvec2& inV1, const vgvec2& inV2) { return(inV1.Length2() < (inV2.Length2()-VG_EPSILON2)); }
VG_CALL bool operator > (const vgvec2& inV1, const vgvec2& inV2) { return(inV1.Length2() > (inV2.Length2()+VG_EPSILON2)); }


//----------------------------------------
//
// vgvec3
// 3D vector/point
//
//----------------------------------------

VG_CALL vgvec3::vgvec3() {}
VG_CALL vgvec3::vgvec3(const vgvec3& inV) { x = inV.x; y = inV.y; z = inV.z; }	
VG_CALL vgvec3::vgvec3(float inX, float inY, float inZ) { x = inX; y = inY; z = inZ; }
VG_CALL vgvec3::vgvec3(const vgvec2& inV) { x = inV.x; y = inV.y; z = 0.0; }

VG_CALL vgvec3& vgvec3::operator = (const vgvec3& inV) { x = inV.x; y = inV.y; z = inV.z; return(*this); }
VG_CALL vgvec3& vgvec3::operator += (const vgvec3& inV) { x += inV.x; y += inV.y; z += inV.z; return(*this); }
VG_CALL vgvec3& vgvec3::operator -= (const vgvec3& inV) { x -= inV.x; y -= inV.y; z -= inV.z; return(*this); }
VG_CALL vgvec3& vgvec3::operator *= (const vgvec3& inV) { x *= inV.x; y *= inV.y; z *= inV.z; return(*this); }
VG_CALL vgvec3& vgvec3::operator /= (const vgvec3& inV) { x /= inV.x; y /= inV.y; z /= inV.z; return(*this); }
VG_CALL vgvec3& vgvec3::operator *= (float inScale) { x *= inScale; y *= inScale; z *= inScale; return(*this); }
VG_CALL vgvec3& vgvec3::operator /= (float inScale) { x /= inScale; y /= inScale; z /= inScale; return(*this); }
VG_CALL vgvec3& vgvec3::operator >>= (const vgframe3& inF) { *this = *this >> inF; return(*this); }
VG_CALL vgvec3& vgvec3::operator <<= (const vgframe3& inF) { *this = *this << inF; return(*this); }
VG_CALL vgvec3& vgvec3::operator >>= (const vgocs3& inOCS) { *this = *this >> inOCS; return(*this); }
VG_CALL vgvec3& vgvec3::operator <<= (const vgocs3& inOCS) { *this = *this << inOCS; return(*this); }

VG_CALL vgvec3::operator float* () { return((float*)this); }

VG_CALL float vgvec3::Length2() const { return(x*x + y*y + z*z); }
VG_CALL float vgvec3::Length() const { return(sqrt(x*x + y*y + z*z)); }
VG_CALL float vgvec3::Normalize() { float a(Length()); float b(1.0/a); x *= b; y *= b; z *= b; return(a); }
VG_CALL int vgvec3::Dominant()
{
	int d(VG_Fabs(y) > VG_Fabs(x));
	if (VG_Fabs(z) > VG_Fabs((*this)[d]))
		d=2;
	return(d);
}

VG_CALL vgvec3 operator - (const vgvec3& inV) { return(vgvec3(-inV.x, -inV.y, -inV.z)); }
VG_CALL vgvec3 operator + (const vgvec3& inV1, const vgvec3& inV2) { return(vgvec3(inV1.x+inV2.x, inV1.y+inV2.y, inV1.z+inV2.z)); }
VG_CALL vgvec3 operator - (const vgvec3& inV1, const vgvec3& inV2) { return(vgvec3(inV1.x-inV2.x, inV1.y-inV2.y, inV1.z-inV2.z)); }
VG_CALL vgvec3 operator * (const vgvec3& inV1, const vgvec3& inV2) { return(vgvec3(inV1.x*inV2.x, inV1.y*inV2.y, inV1.z*inV2.z)); }
VG_CALL vgvec3 operator / (const vgvec3& inV1, const vgvec3& inV2) { return(vgvec3(inV1.x/inV2.x, inV1.y/inV2.y, inV1.z/inV2.z)); }
VG_CALL vgvec3 operator * (const vgvec3& inV, float inScale) { return(vgvec3(inV.x*inScale, inV.y*inScale, inV.z*inScale)); }
VG_CALL vgvec3 operator / (const vgvec3& inV, float inScale) { return(vgvec3(inV.x/inScale, inV.y/inScale, inV.z/inScale)); }
VG_CALL float operator & (const vgvec3& inV1, const vgvec3& inV2) { return((inV2 - inV1).Length()); }
VG_CALL float operator | (const vgvec3& inV1, const vgvec3& inV2) { return(inV1.x*inV2.x + inV1.y*inV2.y + inV1.z*inV2.z); }
VG_CALL vgvec3 operator ^ (const vgvec3& inV1, const vgvec3& inV2)
{
	return(vgvec3(inV1.y*inV2.z-inV1.z*inV2.y, inV1.z*inV2.x-inV1.x*inV2.z, inV1.x*inV2.y-inV1.y*inV2.x));
}
VG_CALL bool operator ! (const vgvec3& inV) { return(inV.Length2() <= VG_EPSILON2); }
VG_CALL bool operator == (const vgvec3& inV1, const vgvec3& inV2) { return((inV2-inV1).Length2() <= VG_EPSILON2); }
VG_CALL bool operator != (const vgvec3& inV1, const vgvec3& inV2) { return((inV2-inV1).Length2() > VG_EPSILON2); }
VG_CALL bool operator <= (const vgvec3& inV1, const vgvec3& inV2) { return(inV1.Length2() <= (inV2.Length2()+VG_EPSILON2)); }
VG_CALL bool operator >= (const vgvec3& inV1, const vgvec3& inV2) { return(inV1.Length2() >= (inV2.Length2()-VG_EPSILON2)); }
VG_CALL bool operator < (const vgvec3& inV1, const vgvec3& inV2) { return(inV1.Length2() < (inV2.Length2()-VG_EPSILON2)); }
VG_CALL bool operator > (const vgvec3& inV1, const vgvec3& inV2) { return(inV1.Length2() > (inV2.Length2()+VG_EPSILON2)); }


//----------------------------------------
//
// vgframe3
// 3D axial frame
//
//----------------------------------------

VG_CALL vgframe3::vgframe3() : vX(1,0,0), vY(0,1,0), vZ(0,0,1) {}
VG_CALL vgframe3::vgframe3(const vgframe3& inF) { vX = inF.vX; vY = inF.vY; vZ = inF.vZ; }
VG_CALL vgframe3::vgframe3(const vgvec3& inX, const vgvec3& inY, const vgvec3& inZ) { vX = inX; vY = inY; vZ = inZ; }
VG_CALL vgframe3::vgframe3(const vgvec3& inAxis, float inRadians, bool normalizeAxis)
{        
    float halfAngle = inRadians*0.5;
    float w = cos(halfAngle); // quaternion scalar
    vgvec3 v = -inAxis;
    if (normalizeAxis)
        v.Normalize();
    v *= sin(halfAngle); // quaternion vector

    float x(v.x), y(v.y), z(v.z);
	float x2(x*2.0), y2(y*2.0), z2(z*2.0), w2(w*2.0);
	float xx2(x*x2), yy2(y*y2), zz2(z*z2), ww2(w*w2);
	float xy2(x*y2), xz2(x*z2), xw2(x*w2), yz2(y*z2), yw2(y*w2), zw2(z*w2);

    // extract frame from quaternion rotation transform
    vX = vgvec3(1.0-(yy2+zz2), xy2+zw2, xz2-yw2);
	vY = vgvec3(xy2-zw2, 1.0-(xx2+zz2), yz2+xw2);
	vZ = vgvec3(xz2+yw2, yz2-xw2, 1.0-(xx2+yy2));
}

VG_CALL vgframe3& vgframe3::operator = (const vgframe3& inF) { vX = inF.vX; vY = inF.vY; vZ = inF.vZ; return(*this); }
VG_CALL vgframe3& vgframe3::operator >>= (const vgframe3& inF) { *this = *this >> inF; return(*this); }
VG_CALL vgframe3& vgframe3::operator <<= (const vgframe3& inF) { *this = *this << inF; return(*this); }

VG_CALL vgvec3 operator >> (const vgvec3& inV, const vgframe3& inF)
{
	return(vgvec3(
		(inV|inF.vX),
		(inV|inF.vY),
		(inV|inF.vZ)
		));
}

VG_CALL vgvec3 operator << (const vgvec3& inV, const vgframe3& inF)
{
	return(vgvec3(
		(inV.x*inF.vX.x + inV.y*inF.vY.x + inV.z*inF.vZ.x),
		(inV.x*inF.vX.y + inV.y*inF.vY.y + inV.z*inF.vZ.y),
		(inV.x*inF.vX.z + inV.y*inF.vY.z + inV.z*inF.vZ.z)
		));
}

VG_CALL vgframe3 operator >> (const vgframe3& inF1, const vgframe3& inF2)
{
	return(vgframe3(inF1.vX >> inF2, inF1.vY >> inF2, inF1.vZ >> inF2));
}

VG_CALL vgframe3 operator << (const vgframe3& inF1, const vgframe3& inF2)
{
	return(vgframe3(inF1.vX << inF2, inF1.vY << inF2, inF1.vZ << inF2));
}


//----------------------------------------
//
// vgocs3
// 3D orthogonal coordinate system
//
//----------------------------------------

VG_CALL vgocs3::vgocs3() : translate(0,0,0), frame(), scale(1,1,1) {}
VG_CALL vgocs3::vgocs3(const vgocs3& inOCS) { translate = inOCS.translate; frame = inOCS.frame; scale = inOCS.scale; }
VG_CALL vgocs3::vgocs3(const vgvec3& inTranslate, const vgframe3& inFrame, const vgvec3& inScale)
{
	frame = inFrame; translate = inTranslate; scale = inScale;
}

VG_CALL vgocs3& vgocs3::operator = (const vgocs3& inOCS)
{
	translate = inOCS.translate; frame = inOCS.frame; scale = inOCS.scale;
	return(*this);
}
VG_CALL vgocs3& vgocs3::operator += (const vgvec3& inTranslate) { translate += inTranslate; return(*this); }
VG_CALL vgocs3& vgocs3::operator -= (const vgvec3& inTranslate) { translate -= inTranslate; return(*this); }	
VG_CALL vgocs3& vgocs3::operator *= (const vgvec3& inScale) { scale *= inScale; return(*this); }
VG_CALL vgocs3& vgocs3::operator /= (const vgvec3& inScale) { scale /= inScale; return(*this); }
VG_CALL vgocs3& vgocs3::operator *= (float inScale) { scale *= inScale; return(*this); }
VG_CALL vgocs3& vgocs3::operator /= (float inScale) { scale /= inScale; return(*this); }
VG_CALL vgocs3& vgocs3::operator >>= (const vgocs3& inOCS) { *this = *this >> inOCS; return(*this); }
VG_CALL vgocs3& vgocs3::operator <<= (const vgocs3& inOCS) { *this = *this << inOCS; return(*this); }

VG_CALL vgvec3 operator >> (const vgvec3& inV, const vgocs3& inOCS)
{
	return(((inV-inOCS.translate) >> inOCS.frame) / inOCS.scale);
}

VG_CALL vgvec3 operator << (const vgvec3& inV, const vgocs3& inOCS)
{
	return(((inV*inOCS.scale) << inOCS.frame) + inOCS.translate);
}

VG_CALL vgocs3 operator >> (const vgocs3& inOCS1, const vgocs3& inOCS2)
{
	return(vgocs3(inOCS1.translate >> inOCS2, inOCS1.frame >> inOCS2.frame, inOCS1.scale / inOCS2.scale));	
}

VG_CALL vgocs3 operator << (const vgocs3& inOCS1, const vgocs3& inOCS2)
{
	return(vgocs3(inOCS1.translate << inOCS2, inOCS1.frame << inOCS2.frame, inOCS1.scale * inOCS2.scale));
}

//============================================================================
//    INTERFACE TRAILING HEADERS
//============================================================================

//****************************************************************************
//**
//**    END HEADER VG_VEC.H
//**
//****************************************************************************
#endif // __VG_VEC_H__
